-- $Id: bwaa-aliasing.ads 105 2009-06-12 15:52:52Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Strings.Unbounded.Hash;

with Ast2Cfg.Flow_Types;
with Ast2Cfg.Pkgs;
with Ast2Cfg.Utils;

with Bwaa.Hashed_Union_Find;
with Bwaa.Util;
with Bwaa.Visibility; use Bwaa.Visibility;
with Bwaa.Visibility.Visibility_Info;

package Bwaa.Aliasing is

   type Alias_Info is limited private;
   type Alias_Info_Ptr is access all Alias_Info;

   -- clear the contents of the alias info
   procedure Clear (Aliases: in out Alias_Info);

   -- free allocated memory, set Alias_Info_Ptr null
   procedure Free (Aliases_Ptr: in out Alias_Info_Ptr);

   -- collect all variables and their aliases starting at the given flow object by going up the pkg/cfg tree
   procedure Collect (Aliases: in out Alias_Info;
                      Flow: in Ast2Cfg.Flow_Types.Flow_Object'Class;
                      Specs: in Bwaa.Util.Specification_Map.Map;
                      Visi_Info: in Visibility_Info.Info);

   -- collect all variables and their aliases from the given flow object
   procedure Collect_From_Flow (Aliases: in out Alias_Info;
                                Specs: in Bwaa.Util.Specification_Map.Map;
                                Visi_Info: in Visibility_Info.Info;
                                Flow: in Ast2Cfg.Flow_Types.Flow_Object'Class);

   -- collect all variables and their aliases from the specifications in the subtree
   procedure Collect_From_Spec_Subtree (Aliases: in out Alias_Info;
                                        Spec: in Ast2Cfg.Pkgs.Spec_Object'Class;
                                        Specs: in Bwaa.Util.Specification_Map.Map;
                                        Root_Name: in Unbounded_String;
                                        Visi_Info: in Visibility_Info.Info);

   -- collect all variables and their aliases from the specifications included by a with clause
   procedure Collect_From_With_Subtree (Aliases: in out Alias_Info;
                                        Flow: in Ast2Cfg.Flow_Types.Flow_Object'Class;
                                        Specs: in Bwaa.Util.Specification_Map.Map;
                                        Visi_Info: in Visibility_Info.Info);

   -- completes a not fully qualified variable name e.g: Bar.X => Foo.Bar.X
   procedure Complete_Prefix(Var_Name: in out Unbounded_String;
                             Flow: in Ast2Cfg.Flow_Types.Flow_Object'Class;
                             Specs: in Bwaa.Util.Specification_Map.Map);

   -- check whether both variable names represent the same variable
   function Equal_Variable_Names (Aliases_Ptr: in Alias_Info_Ptr; X: in Unbounded_String; Y: in Unbounded_String) return Boolean;

   -- debug image of aliases
   function Debug_Image (Aliases: in Alias_Info_Ptr) return String;

   Wrong_Flow_Object_Exception: exception;
   Invalid_Pkg_Cfg_Tree: exception;
   Package_Not_Found: exception;

private

   package USet renames Bwaa.Util.Unbounded_String_Set;

   package Alias_Sets is new Bwaa.Hashed_Union_Find(Unbounded_String, Null_Unbounded_String, Hash);
   use Alias_Sets;

   procedure Add_Prefixed_Series (Alias_Set: in out Alias_Sets.Sets;
                                  Aliases: in out Alias_Info;
                                  X: in Unbounded_String;
                                  Prefix_List: in out Ast2Cfg.Utils.UStr_List.Object;
                                  Lowest_Prefix: in Unbounded_String := Null_Unbounded_String);

   procedure Add_Prefixed_Series (Alias_Set: in out Alias_Sets.Sets;
                                  Aliases: in out Alias_Info;
                                  X: in Unbounded_String;
                                  Prefix_List: in out Ast2Cfg.Utils.UStr_List.Object;
                                  Just_Added: in out USet.Set;
                                  Lowest_Prefix: in Unbounded_String := Null_Unbounded_String);

   procedure Add_Variable_Alias (Aliases: in out Alias_Info; Old_Name: in Unbounded_String; New_Name: in Unbounded_String);
   procedure Add_Package_Alias (Aliases: in out Alias_Info; Old_Name: in Unbounded_String; New_Name: in Unbounded_String);
   procedure Update_Variable_Aliases (Aliases: in out Alias_Info; Old_Pkg_Name: in Unbounded_String; New_Pkg_Name: in Unbounded_String);
   procedure Update_Package_Aliases (Aliases: in out Alias_Info; Old_Pkg_Name: in Unbounded_String; New_Pkg_Name: in Unbounded_String);

   -- check whether both package names represent the same package
   procedure Equal_Package_Names (Aliases: in out Alias_Info; X: in Unbounded_String; Y: in Unbounded_String; Result: out Boolean);

   procedure Circular_Package_Names (Aliases: in out Alias_Info; Old_Name: in Unbounded_String; New_Name: in Unbounded_String; Result: out Boolean);

   type Alias_Info is
      record
         Variable_Aliases: Alias_Sets.Sets;
         Package_Aliases: Alias_Sets.Sets;
      end record;

end Bwaa.Aliasing;
