-- $Id: bwaa-hashed_union_find.ads 98 2009-06-05 21:14:55Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Containers.Hashed_Maps;

-- a union find based data structure with path compression and
-- union-by-rank, the number of elements is dynamic and the implementation
-- uses hashes

generic

   type Element_Type is private;
   Null_Element: in Element_Type;
   with function Hash (Elem: Element_Type) return Ada.Containers.Hash_Type;
   with function "=" (Left, Right : Element_Type) return Boolean is <>;

package Bwaa.Hashed_Union_Find is

   type Sets is private;
   type Cursor is private;

   No_Element: constant Cursor;

   Hashed_Union_Find_Exception: exception;
   Not_In_Any_Set_Exception: exception;


   -- CLASSIC UNION FIND METHODS

   -- create a set containing only X
   procedure Makeset (S: in out Sets; X: in Element_Type);

   -- finds the canonical element for the set containing X; uses path compression
   procedure Find (S: in out Sets; X: in Element_Type; Canonical: out Element_Type);

   -- links X to Y, so that find on X and all its descendants returns
   -- the canonical element of the set containing Y; uses union-by-rank
   procedure Link (S: in out Sets; X: in Element_Type; Y: in Element_Type);

   -- links the canonical elements of the sets containing X or Y
   procedure Union (S: in out Sets; X: in Element_Type; Y: in Element_Type);


   -- ITERATOR METHODS

   -- get a cursor on the first element in the set containing the given element
   procedure First (S: in out Sets; X: in Element_Type; Cur: out Cursor);

   -- get a cursor on the first of all elements
   function First (S: in Sets) return Cursor;

   -- get a cursor on the next element of all elements
   function Next (Position: in Cursor) return Cursor;

   -- change the cursor to point on the next element of all elements
   procedure Next (Position: in out Cursor);

   -- change the cursor to point on the next element in the same set
   procedure Next (S: in out Sets; Position: in out Cursor);

   -- check whether there is an element at the cursor position
   function Has_Element (Position: Cursor) return Boolean;

   -- return the element at the cursor position
   function Element (Position: in Cursor) return Element_Type;

   -- return the parent of the element at the cursor position (for debugging purposes)
   function Parent (Position: in Cursor) return Element_Type;


   -- METHODS USING ALL ELEMENTS AT ONCE

   -- return the number of all elements in all sets
   function Length (S: in Sets) return Ada.Containers.Count_Type;

   -- check whether the element is in one of the sets
   function Contains (S: in Sets; X: in Element_Type) return Boolean;

   -- remove all elements
   procedure Clear (S: in out Sets);

private

   type Rank is new Natural;

   package Element_Map is new Ada.Containers.Hashed_Maps(Key_Type => Element_Type,
                                                         Element_Type => Element_Type,
                                                         Hash => Hash,
                                                         Equivalent_Keys => "=");

   package Rank_Map is new Ada.Containers.Hashed_Maps(Key_Type => Element_Type,
                                                      Element_Type => Rank,
                                                      Hash => Hash,
                                                      Equivalent_Keys => "=");

   type Cursor is
      record
         Cur: Element_Map.Cursor := Element_Map.No_Element;
         Canonical: Element_Type := Null_Element;
      end record;

   No_Element: constant Cursor := (Cur => Element_Map.No_Element, Canonical => Null_Element);

   type Sets is
      record
         Elements: Element_Map.Map;
         Ranks: Rank_Map.Map;
      end record;

end Bwaa.Hashed_Union_Find;
