-- $Id: bwaa-visibility-visibility_info.adb 131 2009-07-09 14:37:32Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ast2Cfg;
with Ast2Cfg.Declaration_Id;
with Ast2cfg.Pkgs;
with Ast2cfg.CFGs;
use type Ast2Cfg.Long_Long_Natural;

with Bwaa.Output; use Bwaa.Output;
with Bwaa.Util; use Bwaa.Util;

package body Bwaa.Visibility.Visibility_Info is

   procedure Try_At_Pred (Full_Old_Var_Name: in out Unbounded_String;
                          Flow: in Ast2Cfg.Flow_Types.Flow_Object'Class;
                          Specs: in Bwaa.Util.Specification_Map.Map;
                          Vm_Specs: in Visibility_Map.Map;
                          Vm_Bodies: in Visibility_Map.Map) is

      use Flow_Class_Ptr_List;
      Full_Flow_Name: Unbounded_String := Get_Full_Name(Flow);
      Preds: Flow_Class_Ptr_List.Object := Get_Preds(Flow);
      Pred_Count: Ast2Cfg.Long_Long_Natural := Count(Preds);
      Flow_Decl_Id: Ast2Cfg.Declaration_Id.Object := Ast2Cfg.Declaration_Id.Null_Declaration_Id;
      Old_Var_Name: Unbounded_String := Full_Old_Var_Name;
   begin

      -- in case this is a body also search in the pred of its spec
      declare
         Spec: Flow_Class_Ptr := null;
         Preds: Flow_Class_Ptr_List.Object;
         Pred_Count: Ast2Cfg.Long_Long_Natural := 0;
      begin
         if Flow in Ast2Cfg.Pkgs.Body_Object'Class or else
            Flow in Ast2Cfg.CFGs.Body_Object'Class then

            Spec := Bwaa.Util.Specification_Map.Element(Specs, Full_Flow_Name);
            Preds := Get_Preds(Spec.all);
            Pred_Count := Count(Preds);

            if Pred_Count > 1 then
               raise Visibility_Exception with "ERROR: Encountered flow with more than one predecessor.";
            elsif Pred_Count = 1 then

               if Spec.all in Ast2Cfg.Pkgs.Pkg_Object'Class then
                  Flow_Decl_Id := Ast2Cfg.Pkgs.Get_Declaration_Id(Ast2cfg.Pkgs.Pkg_Object(Spec.all));
               elsif Spec.all in Ast2Cfg.CFGs.CFG_Object'Class then
                  Flow_Decl_Id := Ast2Cfg.CFGs.Get_Declaration_Id(Ast2cfg.CFGs.CFG_Object(Spec.all));
               else
                  raise Visibility_Exception with "ERROR: Encountered flow that is neither Pkg nor CFG.";
               end if;

               Get_Full_Visible_Name(Full_Old_Var_Name, Flow_Decl_Id, Get_Head(Preds).all, Specs, Vm_Specs, Vm_Bodies, Last_Try => True);
            end if;
         end if;
      exception
         when Constraint_Error =>
            Put_Line(Verbose, "[Bwaa.Visibility.Visibility_Info] Unable to retrieve spec for " & U2s(Full_Flow_Name));
      end;

      if Full_Old_Var_Name /= Old_Var_Name then
         return;
      end if;

      -- search in the pred of the flow
      if Pred_Count < 1 then
         Put_Line(Warning, "[Bwaa.Visibility.Visibility_Info] Warning: Cannot find a variable/package " & U2s(Full_Old_Var_Name) & " that is visible in " & U2s(Full_Flow_Name) & ".");
      elsif Pred_Count > 1 then
         raise Visibility_Exception with "ERROR: Encountered flow with more than one predecessor.";
      else
         if Flow in Ast2Cfg.Pkgs.Pkg_Object'Class then
            Flow_Decl_Id := Ast2Cfg.Pkgs.Get_Declaration_Id(Ast2cfg.Pkgs.Pkg_Object(Flow));
         elsif Flow in Ast2Cfg.CFGs.CFG_Object'Class then
            Flow_Decl_Id := Ast2Cfg.CFGs.Get_Declaration_Id(Ast2cfg.CFGs.CFG_Object(Flow));
         else
            raise Visibility_Exception with "ERROR: Encountered flow that is neither Pkg nor CFG.";
         end if;

         Get_Full_Visible_Name(Full_Old_Var_Name, Flow_Decl_Id, Get_Head(Preds).all, Specs, Vm_Specs, Vm_Bodies, Last_Try => True);
         if Full_Old_Var_Name = Old_Var_Name then
            Put_Line(Warning, "[Bwaa.Visibility.Visibility_Info] Warning: Cannot find a variable/package " & U2s(Full_Old_Var_Name) & " that is visible in " & U2s(Full_Flow_Name) & ".");
         end if;
      end if;
   end Try_At_Pred;

   procedure Get_Full_Visible_Name(Full_Old_Var_Name: in out Unbounded_String;
                                   Decl_Id: in Ast2Cfg.Declaration_Id.Object;
                                   Flow: in Ast2Cfg.Flow_Types.Flow_Object'Class;
                                   Specs: in Bwaa.Util.Specification_Map.Map;
                                   Vm_Specs: in Visibility_Map.Map;
                                   Vm_Bodies: in Visibility_Map.Map;
                                   Last_Try: in Boolean := False) is

      Full_Flow_Name: Unbounded_String := Get_Full_Name(Flow);
      Vm: Visibility_Map.Map := Visibility_Map.Empty_Map;
      Vs: Visible_Set.Set := Visible_Set.Empty_Set;
      Decl: Declaration;
   begin

      Put_Line(Max, "[Bwaa.Visibility.Visibility_Info] Get_Full_Visible_Name called for " & U2s(Full_Old_Var_Name) & " in " & U2s(Full_Flow_Name));

      if Flow in Ast2Cfg.Pkgs.Spec_Object'Class or else
         Flow in Ast2Cfg.CFGs.Spec_Object'Class then

         Vm := Vm_Specs;
      elsif Flow in Ast2Cfg.Pkgs.Body_Object'Class or else
            Flow in Ast2Cfg.CFGs.Body_Object'Class then

         Vm := Vm_Bodies;
      else
         raise Visibility_Exception with "[Bwaa.Visibility.Visibility_Info] ERROR: Encountered flow that is neither spec nor body.";
      end if;

      declare
      begin
         Init(Decl, Full_Old_Var_Name, Null_Unbounded_String, Ast2Cfg.Declaration_Id.Null_Declaration_Id);
      exception
         when Visibility_Exception =>
            null;
      end;

      Vs := Visibility_Map.Element(Vm, Full_Flow_Name);
      Decl := Visible_Set.Element(Visible_Set.Find(Vs, Decl));

      if Decl < Decl_Id or
         (Get_Prefix(Decl) = Get_Prefix(Flow)) then

         Put_Line(Max, "[Bwaa.Visibility.Visibility_Info] " & U2s(Full_Old_Var_Name) & " => " & U2s(Get_Full_Name(Decl)));
         Full_Old_Var_Name := Get_Full_Name(Decl);
      elsif Last_Try then
         null;
      else
         Try_At_Pred(Full_Old_Var_Name, Flow, Specs, Vm_Specs, Vm_Bodies);
      end if;

   exception
      when Constraint_Error =>
         if Last_Try then
            null;
         else
            Try_At_Pred(Full_Old_Var_Name, Flow, Specs, Vm_Specs, Vm_Bodies);
         end if;
   end Get_Full_Visible_Name;

end Bwaa.Visibility.Visibility_Info;
